#include <stdlib.h>
#include <stdio.h>
#include "aswm.h"
#include "event_handlers/create.h"
#include "event_handlers/configure.h"
#include "event_handlers/map.h"
#include "event_handlers/unmap.h"
#include "event_handlers/reparent.h"
#include "event_handlers/destroy.h"
#include "log/error.h"

Aswm _aswm;


void aswm_open() {
	_aswm.display = XOpenDisplay(NULL);
	_aswm.workspaces = malloc(0);
	_aswm.workspace_count = 0;

	XSetErrorHandler(aswm_error_handler);

	Window x_root_window = DefaultRootWindow(_aswm.display);
	XWindowAttributes x_root_attributes;
	XGetWindowAttributes(_aswm.display, x_root_window, &x_root_attributes);

	_aswm.root_window = XCreateSimpleWindow(
			_aswm.display, x_root_window,
			0, 0,
			x_root_attributes.width,
			x_root_attributes.height,
			0, 0, 0);

	// User windows are created from the X root window, while ASWM specific
	// windows are created on the ASWM root window
	XSelectInput(_aswm.display, _aswm.root_window, FocusChangeMask | EnterWindowMask | LeaveWindowMask | ButtonPressMask | ButtonReleaseMask);

	XSelectInput(_aswm.display, x_root_window, SubstructureRedirectMask | SubstructureNotifyMask | FocusChangeMask | EnterWindowMask | LeaveWindowMask | ButtonPressMask | ButtonReleaseMask);

	XMapWindow(_aswm.display, _aswm.root_window);
	XSync(_aswm.display, 0);

	printf("PointerRoot %lu\n", PointerRoot);
	printf("Useful constants:\n");
	printf("\tNone: %lu\n", None);
	printf("\tTrue / False: %i / %i\n", True, False);
	printf("X Root window %lu\n", x_root_window);
	printf("ASWM Root window %lu\n", _aswm.root_window);

	atexit(aswm_close);
}

void aswm_create_workspace() {
	_aswm.workspaces = realloc(_aswm.workspaces, (_aswm.workspace_count+1) * sizeof(Window));

	XWindowAttributes root_attributes;
	XGetWindowAttributes(_aswm.display, _aswm.root_window, &root_attributes);

	_aswm.workspaces[_aswm.workspace_count] = XCreateSimpleWindow(
			_aswm.display, _aswm.root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);

	_aswm.current_workspace = _aswm.workspaces[_aswm.workspace_count];
	_aswm.workspace_count++;
	XSelectInput(_aswm.display, _aswm.current_workspace, SubstructureNotifyMask);
	XMapWindow(_aswm.display, _aswm.current_workspace);
	printf("Create workspace %lu\n", _aswm.current_workspace);
}

void aswm_event_loop() {
	for (;;) {
		XEvent e;
		XNextEvent(_aswm.display, &e);

		switch(e.type) {
			case CreateNotify:
				// Windows with override_redirect set to True can be safely
				// ignored, according to the X server specification.
				if(e.xcreatewindow.override_redirect == False)
					OnCreateNotify(&_aswm, &e.xcreatewindow);
				break;
			case ConfigureRequest:
				OnConfigureRequest(&e.xconfigurerequest);
				break;
			case MapRequest:
				OnMapRequest(&_aswm, &e.xmaprequest);
				break;
			case UnmapNotify:
				OnUnmapNotify(&_aswm, &e.xunmap);
				break;
			case DestroyNotify:
				OnDestroyNotify(&e.xdestroywindow);
				break;
			case ReparentNotify:
				OnReparentNotify(&e.xreparent);
				break;
			case FocusIn:
				printf("FocusIn %lu\n", (&e.xfocus)->window);
				break;
			case FocusOut:
				printf("FocusOut %lu\n", e.xfocus.window);
				break;
			case EnterNotify:
				int revert_to;
				Window window;
				XGetInputFocus(e.xcrossing.display, &window, &revert_to);
				printf("Enter %lu\n", e.xcrossing.window);
				printf("Input focus %lu\n", window);
				break;
			case LeaveNotify:
				printf("Leave %lu\n", e.xcrossing.window);
				break;
			case ButtonPress:
				printf("Press %lu\n", e.xbutton.window);
				break;
			case ButtonRelease:
				printf("Release %lu\n", e.xbutton.window);
				break;
			case ClientMessage:
				// TODO: handle this
				printf("Unhandled client message from window %lu:\n", e.xclient.window);
				char* atom_name = XGetAtomName(e.xclient.display, e.xclient.message_type);
				printf("\t%s\n", atom_name);
				XFree(atom_name);
				break;
			case PropertyNotify:
				// TODO: handle this
				printf("Unhandled property change for window %lu:\n", e.xclient.window);
				char* property_name = XGetAtomName(e.xclient.display, e.xclient.message_type);
				printf("\t%s\n", property_name);
				XFree(property_name);
			default:
				break;
					/* LOG(WARNING) << "Ignored event"; */
		}
	}
}

void aswm_close() {
	free(_aswm.workspaces);
	XCloseDisplay(_aswm.display);
}

int main(int argc, char** argv) {
	aswm_open();
	aswm_event_loop();
	aswm_close();
}
